@@ -1,26 +1,22 @@
-require 'test/unit'
-require 'tmpdir'
-
+# frozen_string_literal: false
 begin
   require 'gdbm'
 rescue LoadError
 end
 
 if defined? GDBM
+  require 'test/unit'
+  require 'envutil' unless defined?(EnvUtil)
   require 'tmpdir'
   require 'fileutils'
 
-  class TestGDBM < Test::Unit::TestCase
-    def TestGDBM.uname_s
+  class TestGDBM_RDONLY < Test::Unit::TestCase
+    def TestGDBM_RDONLY.uname_s
       require 'rbconfig'
       case RbConfig::CONFIG['target_os']
       when 'cygwin'
-        require 'Win32API'
-        uname = Win32API.new('cygwin1', 'uname', 'P', 'I')
-        utsname = ' ' * 100
-        raise 'cannot get system name' if uname.call(utsname) == -1
-
-        utsname.unpack('A20' * 5)[0]
+        require 'etc'
+	Etc.uname[:sysname]
       else
         RbConfig::CONFIG['target_os']
       end
@@ -31,7 +27,6 @@ if defined? GDBM
       @tmpdir = Dir.mktmpdir("tmptest_gdbm")
       @prefix = "tmptest_gdbm_#{$$}"
       @path = "#{@tmpdir}/#{@prefix}_"
-      assert_instance_of(GDBM, @gdbm = GDBM.new(@path))
 
       # prepare to make readonly GDBM file
       GDBM.open("#{@tmpdir}/#{@prefix}_rdonly", 0400) {|gdbm|
@@ -40,7 +35,6 @@ if defined? GDBM
       assert_instance_of(GDBM, @gdbm_rdonly = GDBM.new("#{@tmpdir}/#{@prefix}_rdonly", nil))
     end
     def teardown
-      assert_nil(@gdbm.close)
       assert_nil(@gdbm_rdonly.close)
       ObjectSpace.each_object(GDBM) do |obj|
         obj.close unless obj.closed?
@@ -48,6 +42,41 @@ if defined? GDBM
       FileUtils.remove_entry_secure @tmpdir
     end
 
+    def test_delete_rdonly
+      if /^CYGWIN_9/ !~ SYSTEM
+        assert_raise(GDBMError) {
+          @gdbm_rdonly.delete("foo")
+        }
+
+        assert_nil(@gdbm_rdonly.delete("bar"))
+      end
+    end
+  end
+
+  class TestGDBM < Test::Unit::TestCase
+    SYSTEM = TestGDBM_RDONLY::SYSTEM
+
+    def setup
+      @tmpdir = Dir.mktmpdir("tmptest_gdbm")
+      @prefix = "tmptest_gdbm_#{$$}"
+      @path = "#{@tmpdir}/#{@prefix}_"
+      assert_instance_of(GDBM, @gdbm = GDBM.new(@path))
+    end
+    def teardown
+      assert_nil(@gdbm.close)
+      ObjectSpace.each_object(GDBM) do |obj|
+        obj.close unless obj.closed?
+      end
+      begin
+        FileUtils.remove_entry_secure @tmpdir
+      rescue
+        system("fuser", *Dir.entries(@tmpdir).grep(/\A(?!\.\.?\z)/), chdir: @tmpdir)
+      else
+        return
+      end
+      FileUtils.remove_entry_secure @tmpdir
+    end
+
     def check_size(expect, gdbm=@gdbm)
       assert_equal(expect, gdbm.size)
       n = 0
@@ -60,15 +89,6 @@ if defined? GDBM
       end
     end
 
-    def have_fork?
-      begin
-        fork{}
-        true
-      rescue NotImplementedError
-        false
-      end
-    end
-
     def test_s_new_has_no_block
       # GDBM.new ignore the block
       foo = true
@@ -83,7 +103,7 @@ if defined? GDBM
       begin
         assert_instance_of(GDBM, gdbm = GDBM.open("#{@tmpdir}/#{@prefix}"))
         gdbm.close
-        assert_equal(File.stat("#{@tmpdir}/#{@prefix}").mode & 0777, 0666)
+        assert_equal(File.stat("#{@tmpdir}/#{@prefix}").mode & 0777, 0666) unless /mswin|mingw/ =~ RUBY_PLATFORM
         assert_instance_of(GDBM, gdbm = GDBM.open("#{@tmpdir}/#{@prefix}2", 0644))
         gdbm.close
         assert_equal(File.stat("#{@tmpdir}/#{@prefix}2").mode & 0777, 0644)
@@ -92,8 +112,8 @@ if defined? GDBM
       end
     end
     def test_s_open_no_create
-      assert_nil(gdbm = GDBM.open("#{@tmpdir}/#{@prefix}", nil),
-                 "this test is failed on libgdbm 1.8.0")
+      skip "gdbm_open(GDBM_WRITER) is broken on libgdbm 1.8.0" if /1\.8\.0/ =~ GDBM::VERSION
+      assert_nil(gdbm = GDBM.open("#{@tmpdir}/#{@prefix}", nil))
     ensure
       gdbm.close if gdbm
     end
@@ -118,23 +138,33 @@ if defined? GDBM
     def test_s_open_with_block
       assert_equal(GDBM.open("#{@tmpdir}/#{@prefix}") { :foo }, :foo)
     end
+
+    def open_db_child(dbname, *opts)
+      opts = [0644, *opts].map(&:inspect).join(', ')
+      args = [EnvUtil.rubybin, "-rgdbm", "-e", <<-SRC, dbname]
+      STDOUT.sync = true
+      gdbm = GDBM.open(ARGV.shift, #{opts})
+      puts gdbm.class
+      gets
+      SRC
+      IO.popen(args, "r+") do |f|
+        dbclass = f.gets
+        assert_equal("GDBM", dbclass.chomp)
+        yield
+      end
+    end
+
     def test_s_open_lock
-      return unless have_fork?	# snip this test
-      pid = fork() {
-        assert_instance_of(GDBM, gdbm  = GDBM.open("#{@tmpdir}/#{@prefix}", 0644))
-        sleep 2
-      }
-      begin
-        sleep 1
-        assert_raise(Errno::EWOULDBLOCK) {
-          begin
-            assert_instance_of(GDBM, gdbm2 = GDBM.open("#{@tmpdir}/#{@prefix}", 0644))
-          rescue Errno::EAGAIN, Errno::EACCES
-            raise Errno::EWOULDBLOCK
-          end
+      skip "GDBM.open would block when opening already locked gdbm file on platforms without flock and with lockf" if /solaris|aix/ =~ RUBY_PLATFORM
+
+      dbname = "#{@tmpdir}/#{@prefix}"
+
+      open_db_child(dbname) do
+        assert_raise(Errno::EWOULDBLOCK, Errno::EAGAIN, Errno::EACCES) {
+          GDBM.open(dbname, 0644) {|gdbm|
+            assert_instance_of(GDBM, gdbm)
+          }
         }
-      ensure
-        Process.wait pid
       end
     end
 
@@ -154,51 +184,35 @@ if defined? GDBM
 =end
 
     def test_s_open_nolock
-      # gdbm 1.8.0 specific
-      if not defined? GDBM::NOLOCK
-        return
-      end
-      return unless have_fork?	# snip this test
+      dbname = "#{@tmpdir}/#{@prefix}"
 
-      pid = fork() {
-        assert_instance_of(GDBM, gdbm  = GDBM.open("#{@tmpdir}/#{@prefix}", 0644,
-                                                  GDBM::NOLOCK))
-        sleep 2
-      }
-      sleep 1
-      begin
-        gdbm2 = nil
+      open_db_child(dbname, GDBM::NOLOCK) do
         assert_nothing_raised(Errno::EWOULDBLOCK, Errno::EAGAIN, Errno::EACCES) {
-          assert_instance_of(GDBM, gdbm2 = GDBM.open("#{@tmpdir}/#{@prefix}", 0644))
+          GDBM.open(dbname, 0644) {|gdbm2|
+            assert_instance_of(GDBM, gdbm2)
+          }
         }
-      ensure
-        Process.wait pid
-        gdbm2.close if gdbm2
       end
 
-      p Dir.glob("#{@tmpdir}/#{@prefix}*") if $DEBUG
+      STDERR.puts Dir.glob("#{dbname}*") if $DEBUG
 
-      pid = fork() {
-        assert_instance_of(GDBM, gdbm  = GDBM.open("#{@tmpdir}/#{@prefix}", 0644))
-        sleep 2
-      }
-      begin
-        sleep 1
-        gdbm2 = nil
+      # The following test fails on Windows because flock() implementation
+      # is different from Unix.
+      return if /mswin|mingw/ =~ RUBY_PLATFORM
+
+      open_db_child(dbname) do
         assert_nothing_raised(Errno::EWOULDBLOCK, Errno::EAGAIN, Errno::EACCES) {
           # this test is failed on Cygwin98 (???)
-          assert_instance_of(GDBM, gdbm2 = GDBM.open("#{@tmpdir}/#{@prefix}", 0644,
-                                                     GDBM::NOLOCK))
+          GDBM.open(dbname, 0644, GDBM::NOLOCK) {|gdbm2|
+            assert_instance_of(GDBM, gdbm2)
+          }
         }
-      ensure
-        Process.wait pid
-        gdbm2.close if gdbm2
       end
-    end
+    end if defined? GDBM::NOLOCK # gdbm 1.8.0 specific
 
     def test_s_open_error
       assert_instance_of(GDBM, gdbm = GDBM.open("#{@tmpdir}/#{@prefix}", 0))
-      assert_raise(Errno::EACCES) {
+      assert_raise(Errno::EACCES, Errno::EWOULDBLOCK) {
         GDBM.open("#{@tmpdir}/#{@prefix}", 0)
       }
       gdbm.close
@@ -260,7 +274,7 @@ if defined? GDBM
         num += 1 if i == 0
         assert_equal(num, @gdbm.size)
 
-        # Fixnum
+        # Integer
         assert_equal('200', @gdbm['100'] = '200')
         assert_equal('200', @gdbm['100'])
 
@@ -276,9 +290,9 @@ if defined? GDBM
       }
     end
 
-    def test_index
+    def test_key
       assert_equal('bar', @gdbm['foo'] = 'bar')
-      assert_equal('foo', @gdbm.index('bar'))
+      assert_equal('foo', @gdbm.key('bar'))
       assert_nil(@gdbm['bar'])
     end
 
@@ -354,7 +368,7 @@ if defined? GDBM
 
       n = 0
       ret = @gdbm.each_value {|val|
-        assert_not_nil(key = @gdbm.index(val))
+        assert_not_nil(key = @gdbm.key(val))
         assert_not_nil(i = keys.index(key))
         assert_equal(val, values[i])
 
@@ -385,6 +399,10 @@ if defined? GDBM
       assert_equal(@gdbm, ret)
     end
 
+    def test_each_key_without_block
+      assert_kind_of Enumerator, @gdbm.each_key
+    end
+
     def test_keys
       assert_equal([], @gdbm.keys)
 
@@ -438,15 +456,8 @@ if defined? GDBM
       assert_equal(2, @gdbm.size)
 
       assert_nil(@gdbm.delete(key))
-
-      if /^CYGWIN_9/ !~ SYSTEM
-        assert_raise(GDBMError) {
-          @gdbm_rdonly.delete("foo")
-        }
-
-        assert_nil(@gdbm_rdonly.delete("bar"))
-      end
     end
+
     def test_delete_with_block
       key = 'no called block'
       @gdbm[key] = 'foo'
@@ -499,7 +510,7 @@ if defined? GDBM
           n+=1
           true
         }
-      rescue
+      rescue RuntimeError
       end
       assert_equal(51, n)
       check_size(49, @gdbm)
@@ -584,6 +595,7 @@ if defined? GDBM
 
       size2 = File.size(@path)
       @gdbm.reorganize
+      @gdbm.sync
       size3 = File.size(@path)
 
       # p [size1, size2, size3]
@@ -667,6 +679,8 @@ if defined? GDBM
     end
 
     def test_writer_open_notexist
+      skip "gdbm_open(GDBM_WRITER) is broken on libgdbm 1.8.0" if /1\.8\.0/ =~ GDBM::VERSION
+
       assert_raise(Errno::ENOENT) {
         GDBM.open("#{@tmproot}/a", 0666, GDBM::WRITER)
       }
@@ -696,7 +710,7 @@ if defined? GDBM
     def test_newdb_open
       GDBM.open("#{@tmproot}/a.dbm") {|dbm|
         dbm["k"] = "v"
-      } 
+      }
       v = GDBM.open("#{@tmproot}/a.dbm", nil, GDBM::NEWDB) {|d|
         assert_equal(0, d.length)
         assert_nil(d["k"])
