@@ -1,9 +1,9 @@
 # coding: US-ASCII
+# frozen_string_literal: false
 require 'test/unit'
 require 'net/http'
 require 'stringio'
 require_relative 'utils'
-require_relative '../../ruby/envutil'
 
 class TestNetHTTP < Test::Unit::TestCase
 
@@ -14,7 +14,7 @@ class TestNetHTTP < Test::Unit::TestCase
 
     proxy_class = Net::HTTP.Proxy 'proxy.example', 8000, 'user', 'pass'
 
-    refute_equal Net::HTTP, proxy_class
+    assert_not_equal Net::HTTP, proxy_class
 
     assert_operator proxy_class, :<, Net::HTTP
 
@@ -23,9 +23,9 @@ class TestNetHTTP < Test::Unit::TestCase
     assert_equal 'user',          proxy_class.proxy_user
     assert_equal 'pass',          proxy_class.proxy_pass
 
-    http = proxy_class.new 'example'
+    http = proxy_class.new 'hostname.example'
 
-    refute http.proxy_from_env?
+    assert_not_predicate http, :proxy_from_env?
 
 
     proxy_class = Net::HTTP.Proxy 'proxy.example'
@@ -43,7 +43,7 @@ class TestNetHTTP < Test::Unit::TestCase
 
       proxy_class = Net::HTTP.Proxy :ENV
 
-      refute_equal Net::HTTP, proxy_class
+      assert_not_equal Net::HTTP, proxy_class
 
       assert_operator proxy_class, :<, Net::HTTP
 
@@ -51,16 +51,16 @@ class TestNetHTTP < Test::Unit::TestCase
       assert_nil proxy_class.proxy_user
       assert_nil proxy_class.proxy_pass
 
-      refute_equal 8000, proxy_class.proxy_port
+      assert_not_equal 8000, proxy_class.proxy_port
 
-      http = proxy_class.new 'example'
+      http = proxy_class.new 'hostname.example'
 
       assert http.proxy_from_env?
     end
   end
 
   def test_edit_path
-    http = Net::HTTP.new 'example', nil, nil
+    http = Net::HTTP.new 'hostname.example', nil, nil
 
     edited = http.send :edit_path, '/path'
 
@@ -74,11 +74,11 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_edit_path_proxy
-    http = Net::HTTP.new 'example', nil, 'proxy.example'
+    http = Net::HTTP.new 'hostname.example', nil, 'proxy.example'
 
     edited = http.send :edit_path, '/path'
 
-    assert_equal 'http://example/path', edited
+    assert_equal 'http://hostname.example/path', edited
 
     http.use_ssl = true
 
@@ -89,19 +89,40 @@ class TestNetHTTP < Test::Unit::TestCase
 
   def test_proxy_address
     clean_http_proxy_env do
-      http = Net::HTTP.new 'example', nil, 'proxy.example'
+      http = Net::HTTP.new 'hostname.example', nil, 'proxy.example'
       assert_equal 'proxy.example', http.proxy_address
 
-      http = Net::HTTP.new 'example', nil
+      http = Net::HTTP.new 'hostname.example', nil
       assert_equal nil, http.proxy_address
     end
   end
 
+  def test_proxy_address_no_proxy
+    clean_http_proxy_env do
+      http = Net::HTTP.new 'hostname.example', nil, 'proxy.example', nil, nil, nil, 'example'
+      assert_nil http.proxy_address
+
+      http = Net::HTTP.new '10.224.1.1', nil, 'proxy.example', nil, nil, nil, 'example,10.224.0.0/22'
+      assert_nil http.proxy_address
+    end
+  end
+
+  def test_proxy_from_env_ENV
+    clean_http_proxy_env do
+      ENV['http_proxy'] = 'http://proxy.example:8000'
+
+      assert_equal false, Net::HTTP.proxy_class?
+      http = Net::HTTP.new 'hostname.example'
+
+      assert_equal true, http.proxy_from_env?
+    end
+  end
+
   def test_proxy_address_ENV
     clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
-      http = Net::HTTP.new 'example'
+      http = Net::HTTP.new 'hostname.example'
 
       assert_equal 'proxy.example', http.proxy_address
     end
@@ -109,7 +130,7 @@ class TestNetHTTP < Test::Unit::TestCase
 
   def test_proxy_eh_no_proxy
     clean_http_proxy_env do
-      assert_equal false, Net::HTTP.new('example', nil, nil).proxy?
+      assert_equal false, Net::HTTP.new('hostname.example', nil, nil).proxy?
     end
   end
 
@@ -117,35 +138,52 @@ class TestNetHTTP < Test::Unit::TestCase
     clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
-      http = Net::HTTP.new 'example'
+      http = Net::HTTP.new 'hostname.example'
 
       assert_equal true, http.proxy?
     end
   end
 
+  def test_proxy_eh_ENV_with_user
+    clean_http_proxy_env do
+      ENV['http_proxy'] = 'http://foo:bar@proxy.example:8000'
+
+      http = Net::HTTP.new 'hostname.example'
+
+      assert_equal true, http.proxy?
+      if Net::HTTP::ENVIRONMENT_VARIABLE_IS_MULTIUSER_SAFE
+        assert_equal 'foo', http.proxy_user
+        assert_equal 'bar', http.proxy_pass
+      else
+        assert_nil http.proxy_user
+        assert_nil http.proxy_pass
+      end
+    end
+  end
+
   def test_proxy_eh_ENV_none_set
     clean_http_proxy_env do
-      assert_equal false, Net::HTTP.new('example').proxy?
+      assert_equal false, Net::HTTP.new('hostname.example').proxy?
     end
   end
 
   def test_proxy_eh_ENV_no_proxy
     clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
-      ENV['no_proxy']   = 'example'
+      ENV['no_proxy']   = 'hostname.example'
 
-      assert_equal false, Net::HTTP.new('example').proxy?
+      assert_equal false, Net::HTTP.new('hostname.example').proxy?
     end
   end
 
   def test_proxy_port
     clean_http_proxy_env do
-      http = Net::HTTP.new 'exmaple', nil, 'proxy.example'
+      http = Net::HTTP.new 'example', nil, 'proxy.example'
       assert_equal 'proxy.example', http.proxy_address
       assert_equal 80, http.proxy_port
-      http = Net::HTTP.new 'exmaple', nil, 'proxy.example', 8000
+      http = Net::HTTP.new 'example', nil, 'proxy.example', 8000
       assert_equal 8000, http.proxy_port
-      http = Net::HTTP.new 'exmaple', nil
+      http = Net::HTTP.new 'example', nil
       assert_equal nil, http.proxy_port
     end
   end
@@ -154,7 +192,7 @@ class TestNetHTTP < Test::Unit::TestCase
     clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
-      http = Net::HTTP.new 'example'
+      http = Net::HTTP.new 'hostname.example'
 
       assert_equal 8000, http.proxy_port
     end
@@ -164,7 +202,7 @@ class TestNetHTTP < Test::Unit::TestCase
     clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
-      http = Net::HTTP.newobj 'example'
+      http = Net::HTTP.newobj 'hostname.example'
 
       assert_equal false, http.proxy?
     end
@@ -189,10 +227,51 @@ class TestNetHTTP < Test::Unit::TestCase
     end
   end
 
+  def test_failure_message_includes_failed_domain_and_port
+    # hostname to be included in the error message
+    host = Struct.new(:to_s).new("<example>")
+    port = 2119
+    # hack to let TCPSocket.open fail
+    def host.to_str; raise SocketError, "open failure"; end
+    uri = Struct.new(:scheme, :hostname, :port).new("http", host, port)
+    assert_raise_with_message(SocketError, /#{host}:#{port}/) do
+      clean_http_proxy_env{ Net::HTTP.get(uri) }
+    end
+  end
+
 end
 
 module TestNetHTTP_version_1_1_methods
 
+  def test_s_start
+    begin
+      h = Net::HTTP.start(config('host'), config('port'))
+    ensure
+      h&.finish
+    end
+    assert_equal config('host'), h.address
+    assert_equal config('port'), h.port
+    assert_equal true, h.instance_variable_get(:@proxy_from_env)
+
+    begin
+      h = Net::HTTP.start(config('host'), config('port'), :ENV)
+    ensure
+      h&.finish
+    end
+    assert_equal config('host'), h.address
+    assert_equal config('port'), h.port
+    assert_equal true, h.instance_variable_get(:@proxy_from_env)
+
+    begin
+      h = Net::HTTP.start(config('host'), config('port'), nil)
+    ensure
+      h&.finish
+    end
+    assert_equal config('host'), h.address
+    assert_equal config('port'), h.port
+    assert_equal false, h.instance_variable_get(:@proxy_from_env)
+  end
+
   def test_s_get
     assert_equal $test_net_http_data,
         Net::HTTP.get(config('host'), '/', config('port'))
@@ -277,6 +356,7 @@ module TestNetHTTP_version_1_1_methods
       end
     }
     assert_equal 1, i
+    @log_tester = nil # server may encount ECONNRESET
   end
 
   def test_get__implicit_start
@@ -291,6 +371,17 @@ module TestNetHTTP_version_1_1_methods
     assert_equal $test_net_http_data, res.body
   end
 
+  def test_get__crlf
+    start {|http|
+      assert_raise(ArgumentError) do
+        http.get("\r")
+      end
+      assert_raise(ArgumentError) do
+        http.get("\n")
+      end
+    }
+  end
+
   def test_get2
     start {|http|
       http.get2('/') {|res|
@@ -349,6 +440,23 @@ module TestNetHTTP_version_1_1_methods
     end
   end
 
+  def test_s_post
+    url = "http://#{config('host')}:#{config('port')}/?q=a"
+    res = Net::HTTP.post(
+              URI.parse(url),
+              "a=x")
+    assert_equal "application/x-www-form-urlencoded", res["Content-Type"]
+    assert_equal "a=x", res.body
+    assert_equal url, res["X-request-uri"]
+
+    res = Net::HTTP.post(
+              URI.parse(url),
+              "hello world",
+              "Content-Type" => "text/plain; charset=US-ASCII")
+    assert_equal "text/plain; charset=US-ASCII", res["Content-Type"]
+    assert_equal "hello world", res.body
+  end
+
   def test_s_post_form
     url = "http://#{config('host')}:#{config('port')}/"
     res = Net::HTTP.post_form(
@@ -397,13 +505,14 @@ module TestNetHTTP_version_1_1_methods
   def test_timeout_during_HTTP_session
     bug4246 = "expected the HTTP session to have timed out but have not. c.f. [ruby-core:34203]"
 
+    th = nil
     # listen for connections... but deliberately do not read
     TCPServer.open('localhost', 0) {|server|
       port = server.addr[1]
 
       conn = Net::HTTP.new('localhost', port)
       conn.read_timeout = 0.01
-      conn.open_timeout = 0.01
+      conn.open_timeout = 0.1
 
       th = Thread.new do
         assert_raise(Net::ReadTimeout) {
@@ -412,6 +521,9 @@ module TestNetHTTP_version_1_1_methods
       end
       assert th.join(10), bug4246
     }
+  ensure
+    th.kill
+    th.join
   end
 end
 
@@ -460,7 +572,7 @@ module TestNetHTTP_version_1_2_methods
       assert_equal $test_net_http_data.size, res.body.size
       assert_equal $test_net_http_data, res.body
 
-      refute res.decode_content, 'Bug #7831' if Net::HTTP::HAVE_ZLIB
+      assert_not_predicate res, :decode_content, 'Bug #7831' if Net::HTTP::HAVE_ZLIB
     }
   end
 
@@ -527,53 +639,54 @@ module TestNetHTTP_version_1_2_methods
   end
 
   def _test_request__path(http)
-    uri = URI 'https://example/'
+    uri = URI 'https://hostname.example/'
     req = Net::HTTP::Get.new('/')
 
     res = http.request(req)
 
     assert_kind_of URI::Generic, req.uri
 
-    refute_equal uri, req.uri
+    assert_not_equal uri, req.uri
 
     assert_equal uri, res.uri
 
-    refute_same uri,     req.uri
-    refute_same req.uri, res.uri
+    assert_not_same uri,     req.uri
+    assert_not_same req.uri, res.uri
   end
 
   def _test_request__uri(http)
-    uri = URI 'https://example/'
+    uri = URI 'https://hostname.example/'
     req = Net::HTTP::Get.new(uri)
 
     res = http.request(req)
 
     assert_kind_of URI::Generic, req.uri
 
-    refute_equal uri, req.uri
+    assert_not_equal uri, req.uri
 
     assert_equal req.uri, res.uri
 
-    refute_same uri,     req.uri
-    refute_same req.uri, res.uri
+    assert_not_same uri,     req.uri
+    assert_not_same req.uri, res.uri
   end
 
   def _test_request__uri_host(http)
-    uri = URI 'http://example/'
+    uri = URI 'http://other.example/'
 
     req = Net::HTTP::Get.new(uri)
-    req['host'] = 'other.example'
+    req['host'] = 'hostname.example'
 
     res = http.request(req)
 
     assert_kind_of URI::Generic, req.uri
 
-    assert_equal URI("http://example:#{http.port}"), res.uri
+    assert_equal URI("http://hostname.example:#{http.port}"), res.uri
   end
 
   def test_send_request
     start {|http|
       _test_send_request__GET http
+      _test_send_request__HEAD http
       _test_send_request__POST http
     }
   end
@@ -588,6 +701,16 @@ module TestNetHTTP_version_1_2_methods
     assert_equal $test_net_http_data, res.body
   end
 
+  def _test_send_request__HEAD(http)
+    res = http.send_request('HEAD', '/')
+    assert_kind_of Net::HTTPResponse, res
+    unless self.is_a?(TestNetHTTP_v1_2_chunked)
+      assert_not_nil res['content-length']
+      assert_equal $test_net_http_data.size, res['content-length'].to_i
+    end
+    assert_nil res.body
+  end
+
   def _test_send_request__POST(http)
     data = 'aaabbb cc ddddddddddd lkjoiu4j3qlkuoa'
     res = http.send_request('POST', '/', data, 'content-type' => 'application/x-www-form-urlencoded')
@@ -700,7 +823,6 @@ end
 class TestNetHTTP_v1_2 < Test::Unit::TestCase
   CONFIG = {
     'host' => '127.0.0.1',
-    'port' => 0,
     'proxy_host' => nil,
     'proxy_port' => nil,
   }
@@ -718,7 +840,6 @@ end
 class TestNetHTTP_v1_2_chunked < Test::Unit::TestCase
   CONFIG = {
     'host' => '127.0.0.1',
-    'port' => 0,
     'proxy_host' => nil,
     'proxy_port' => nil,
     'chunked' => true,
@@ -749,7 +870,6 @@ end
 class TestNetHTTPContinue < Test::Unit::TestCase
   CONFIG = {
     'host' => '127.0.0.1',
-    'port' => 0,
     'proxy_host' => nil,
     'proxy_port' => nil,
     'chunked' => true,
@@ -814,6 +934,22 @@ class TestNetHTTPContinue < Test::Unit::TestCase
     assert_not_match(/HTTP\/1.1 100 continue/, @debug.string)
   end
 
+  def test_expect_continue_error_before_body
+    @log_tester = nil
+    mount_proc {|req, res|
+      raise WEBrick::HTTPStatus::Forbidden
+    }
+    start {|http|
+      uheader = {'content-length' => '5', 'expect' => '100-continue'}
+      http.continue_timeout = 1 # allow the server to respond before sending
+      http.request_post('/continue', 'data', uheader) {|res|
+        assert_equal(res.code, '403')
+      }
+    }
+    assert_match(/Expect: 100-continue/, @debug.string)
+    assert_not_match(/HTTP\/1.1 100 continue/, @debug.string)
+  end
+
   def test_expect_continue_error_while_waiting
     mount_proc {|req, res|
       res.status = 501
@@ -831,10 +967,42 @@ class TestNetHTTPContinue < Test::Unit::TestCase
   end
 end
 
+class TestNetHTTPSwitchingProtocols < Test::Unit::TestCase
+  CONFIG = {
+    'host' => '127.0.0.1',
+    'proxy_host' => nil,
+    'proxy_port' => nil,
+    'chunked' => true,
+  }
+
+  include TestNetHTTPUtils
+
+  def logfile
+    @debug = StringIO.new('')
+  end
+
+  def mount_proc(&block)
+    @server.mount('/continue', WEBrick::HTTPServlet::ProcHandler.new(block.to_proc))
+  end
+
+  def test_info
+    mount_proc {|req, res|
+      req.instance_variable_get(:@socket) << "HTTP/1.1 101 Switching Protocols\r\n\r\n"
+      res.body = req.query['body']
+    }
+    start {|http|
+      http.continue_timeout = 0.2
+      http.request_post('/continue', 'body=BODY') {|res|
+        assert_equal('BODY', res.read_body)
+      }
+    }
+    assert_match(/HTTP\/1.1 101 Switching Protocols/, @debug.string)
+  end
+end
+
 class TestNetHTTPKeepAlive < Test::Unit::TestCase
   CONFIG = {
     'host' => '127.0.0.1',
-    'port' => 0,
     'proxy_host' => nil,
     'proxy_port' => nil,
     'RequestTimeout' => 1,
@@ -857,6 +1025,23 @@ class TestNetHTTPKeepAlive < Test::Unit::TestCase
     }
   end
 
+  def test_server_closed_connection_auto_reconnect
+    start {|http|
+      res = http.get('/')
+      http.keep_alive_timeout = 5
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+      sleep 1.5
+      assert_nothing_raised {
+        # Net::HTTP should detect the closed connection before attempting the
+        # request, since post requests cannot be retried.
+        res = http.post('/', 'query=foo', 'content-type' => 'application/x-www-form-urlencoded')
+      }
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+    }
+  end
+
   def test_keep_alive_get_auto_retry
     start {|http|
       res = http.get('/')
@@ -870,13 +1055,65 @@ class TestNetHTTPKeepAlive < Test::Unit::TestCase
     }
   end
 
+  class MockSocket
+    attr_reader :count
+    def initialize(success_after: nil)
+      @success_after = success_after
+      @count = 0
+    end
+    def close
+    end
+    def closed?
+    end
+    def write(_)
+    end
+    def readline
+      @count += 1
+      if @success_after && @success_after <= @count
+        "HTTP/1.1 200 OK"
+      else
+        raise Errno::ECONNRESET
+      end
+    end
+    def readuntil(*_)
+      ""
+    end
+    def read_all(_)
+    end
+  end
+
+  def test_http_retry_success
+    start {|http|
+      socket = MockSocket.new(success_after: 10)
+      http.instance_variable_get(:@socket).close
+      http.instance_variable_set(:@socket, socket)
+      assert_equal 0, socket.count
+      http.max_retries = 10
+      res = http.get('/')
+      assert_equal 10, socket.count
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+    }
+  end
+
+  def test_http_retry_failed
+    start {|http|
+      socket = MockSocket.new
+      http.instance_variable_get(:@socket).close
+      http.instance_variable_set(:@socket, socket)
+      http.max_retries = 10
+      assert_raise(Errno::ECONNRESET){ http.get('/') }
+      assert_equal 11, socket.count
+    }
+  end
+
   def test_keep_alive_server_close
     def @server.run(sock)
       sock.close
     end
 
     start {|http|
-      assert_raises(EOFError, Errno::ECONNRESET, IOError) {
+      assert_raise(EOFError, Errno::ECONNRESET, IOError) {
         http.get('/')
       }
     }
@@ -886,7 +1123,6 @@ end
 class TestNetHTTPLocalBind < Test::Unit::TestCase
   CONFIG = {
     'host' => 'localhost',
-    'port' => 0,
     'proxy_host' => nil,
     'proxy_port' => nil,
   }
