@@ -10,7 +10,7 @@
 **********************************************************************/
 
 /*
- * This file incldue two parts:
+ * This file include two parts:
  *
  * (1) set_trace_func internal mechanisms
  *     and C level API
@@ -21,11 +21,9 @@
  *
  */
 
-#include "ruby/ruby.h"
-#include "ruby/debug.h"
-#include "ruby/encoding.h"
-
 #include "internal.h"
+#include "ruby/debug.h"
+
 #include "vm_core.h"
 #include "eval_intern.h"
 
@@ -67,7 +65,7 @@ recalc_add_ruby_vm_event_flags(rb_event_flag_t events)
     ruby_vm_event_flags = 0;
 
     for (i=0; i<MAX_EVENT_NUM; i++) {
-	if (events & (1 << i)) {
+	if (events & ((rb_event_flag_t)1 << i)) {
 	    ruby_event_flag_count[i]++;
 	}
 	ruby_vm_event_flags |= ruby_event_flag_count[i] ? (1<<i) : 0;
@@ -83,7 +81,7 @@ recalc_remove_ruby_vm_event_flags(rb_event_flag_t events)
     ruby_vm_event_flags = 0;
 
     for (i=0; i<MAX_EVENT_NUM; i++) {
-	if (events & (1 << i)) {
+	if (events & ((rb_event_flag_t)1 << i)) {
 	    ruby_event_flag_count[i]--;
 	}
 	ruby_vm_event_flags |= ruby_event_flag_count[i] ? (1<<i) : 0;
@@ -94,18 +92,16 @@ recalc_remove_ruby_vm_event_flags(rb_event_flag_t events)
 
 /* add/remove hooks */
 
-static rb_thread_t *
-thval2thread_t(VALUE thval)
-{
-    rb_thread_t *th;
-    GetThreadPtr(thval, th);
-    return th;
-}
-
 static rb_event_hook_t *
 alloc_event_hook(rb_event_hook_func_t func, rb_event_flag_t events, VALUE data, rb_event_hook_flag_t hook_flags)
 {
-    rb_event_hook_t *hook = ALLOC(rb_event_hook_t);
+    rb_event_hook_t *hook;
+
+    if ((events & RUBY_INTERNAL_EVENT_MASK) && (events & ~RUBY_INTERNAL_EVENT_MASK)) {
+	rb_raise(rb_eTypeError, "Can not specify normal event and internal event simultaneously.");
+    }
+
+    hook = ALLOC(rb_event_hook_t);
     hook->hook_flags = hook_flags;
     hook->events = events;
     hook->func = func;
@@ -132,7 +128,7 @@ rb_threadptr_add_event_hook(rb_thread_t *th, rb_event_hook_func_t func, rb_event
 void
 rb_thread_add_event_hook(VALUE thval, rb_event_hook_func_t func, rb_event_flag_t events, VALUE data)
 {
-    rb_threadptr_add_event_hook(thval2thread_t(thval), func, events, data, RUBY_EVENT_HOOK_FLAG_SAFE);
+    rb_threadptr_add_event_hook(rb_thread_ptr(thval), func, events, data, RUBY_EVENT_HOOK_FLAG_SAFE);
 }
 
 void
@@ -145,7 +141,7 @@ rb_add_event_hook(rb_event_hook_func_t func, rb_event_flag_t events, VALUE data)
 void
 rb_thread_add_event_hook2(VALUE thval, rb_event_hook_func_t func, rb_event_flag_t events, VALUE data, rb_event_hook_flag_t hook_flags)
 {
-    rb_threadptr_add_event_hook(thval2thread_t(thval), func, events, data, hook_flags);
+    rb_threadptr_add_event_hook(rb_thread_ptr(thval), func, events, data, hook_flags);
 }
 
 void
@@ -167,7 +163,7 @@ remove_event_hook(rb_hook_list_t *list, rb_event_hook_func_t func, VALUE data)
 	    if (data == Qundef || hook->data == data) {
 		hook->hook_flags |= RUBY_EVENT_HOOK_FLAG_DELETED;
 		ret+=1;
-		list->need_clean++;
+		list->need_clean = TRUE;
 	    }
 	}
 	hook = hook->next;
@@ -185,13 +181,13 @@ rb_threadptr_remove_event_hook(rb_thread_t *th, rb_event_hook_func_t func, VALUE
 int
 rb_thread_remove_event_hook(VALUE thval, rb_event_hook_func_t func)
 {
-    return rb_threadptr_remove_event_hook(thval2thread_t(thval), func, Qundef);
+    return rb_threadptr_remove_event_hook(rb_thread_ptr(thval), func, Qundef);
 }
 
 int
 rb_thread_remove_event_hook_with_data(VALUE thval, rb_event_hook_func_t func, VALUE data)
 {
-    return rb_threadptr_remove_event_hook(thval2thread_t(thval), func, data);
+    return rb_threadptr_remove_event_hook(rb_thread_ptr(thval), func, data);
 }
 
 int
@@ -206,19 +202,15 @@ rb_remove_event_hook_with_data(rb_event_hook_func_t func, VALUE data)
     return remove_event_hook(&GET_VM()->event_hooks, func, data);
 }
 
-static int
-clear_trace_func_i(st_data_t key, st_data_t val, st_data_t flag)
-{
-    rb_thread_t *th;
-    GetThreadPtr((VALUE)key, th);
-    rb_threadptr_remove_event_hook(th, 0, Qundef);
-    return ST_CONTINUE;
-}
-
 void
 rb_clear_trace_func(void)
 {
-    st_foreach(GET_VM()->living_threads, clear_trace_func_i, (st_data_t) 0);
+    rb_vm_t *vm = GET_VM();
+    rb_thread_t *th = 0;
+
+    list_for_each(&vm->living_threads, th, vmlt_node) {
+	rb_threadptr_remove_event_hook(th, 0, Qundef);
+    }
     rb_remove_event_hook(0);
 }
 
@@ -230,7 +222,7 @@ clean_hooks(rb_hook_list_t *list)
     rb_event_hook_t *hook, **nextp = &list->hooks;
 
     list->events = 0;
-    list->need_clean = 0;
+    list->need_clean = FALSE;
 
     while ((hook = *nextp) != 0) {
 	if (hook->hook_flags & RUBY_EVENT_HOOK_FLAG_DELETED) {
@@ -245,34 +237,57 @@ clean_hooks(rb_hook_list_t *list)
     }
 }
 
-static int
-exec_hooks(rb_thread_t *th, rb_hook_list_t *list, const rb_trace_arg_t *trace_arg, int can_clean_hooks)
+static void
+exec_hooks_body(rb_thread_t *th, rb_hook_list_t *list, const rb_trace_arg_t *trace_arg)
 {
-    int state;
+    rb_event_hook_t *hook;
+
+    for (hook = list->hooks; hook; hook = hook->next) {
+	if (!(hook->hook_flags & RUBY_EVENT_HOOK_FLAG_DELETED) && (trace_arg->event & hook->events)) {
+	    if (!(hook->hook_flags & RUBY_EVENT_HOOK_FLAG_RAW_ARG)) {
+		(*hook->func)(trace_arg->event, hook->data, trace_arg->self, trace_arg->id, trace_arg->klass);
+	    }
+	    else {
+		(*((rb_event_hook_raw_arg_func_t)hook->func))(hook->data, trace_arg);
+	    }
+	}
+    }
+}
+
+static int
+exec_hooks_precheck(rb_thread_t *th, rb_hook_list_t *list, const rb_trace_arg_t *trace_arg)
+{
+    if (UNLIKELY(list->need_clean != FALSE)) {
+	if (th->vm->trace_running <= 1) { /* only running this hooks */
+	    clean_hooks(list);
+	}
+    }
+
+    return (list->events & trace_arg->event) != 0;
+}
+
+static void
+exec_hooks_unprotected(rb_thread_t *th, rb_hook_list_t *list, const rb_trace_arg_t *trace_arg)
+{
+    if (exec_hooks_precheck(th, list, trace_arg) == 0) return;
+    exec_hooks_body(th, list, trace_arg);
+}
+
+static int
+exec_hooks_protected(rb_thread_t *th, rb_hook_list_t *list, const rb_trace_arg_t *trace_arg)
+{
+    enum ruby_tag_type state;
     volatile int raised;
 
-    if (UNLIKELY(list->need_clean > 0) && can_clean_hooks) {
-	clean_hooks(list);
-    }
+    if (exec_hooks_precheck(th, list, trace_arg) == 0) return 0;
 
     raised = rb_threadptr_reset_raised(th);
 
     /* TODO: Support !RUBY_EVENT_HOOK_FLAG_SAFE hooks */
 
     TH_PUSH_TAG(th);
-    if ((state = TH_EXEC_TAG()) == 0) {
-	rb_event_hook_t *hook;
-
-	for (hook = list->hooks; hook; hook = hook->next) {
-	    if (!(hook->hook_flags & RUBY_EVENT_HOOK_FLAG_DELETED) && (trace_arg->event & hook->events)) {
-		if (!(hook->hook_flags & RUBY_EVENT_HOOK_FLAG_RAW_ARG)) {
-		    (*hook->func)(trace_arg->event, hook->data, trace_arg->self, trace_arg->id, trace_arg->klass);
-		}
-		else {
-		    (*((rb_event_hook_raw_arg_func_t)hook->func))(hook->data, trace_arg);
-		}
-	    }
-	}
+    if ((state = TH_EXEC_TAG()) == TAG_NONE) {
+	exec_hooks_body(th, list, trace_arg);
     }
     TH_POP_TAG();
 
@@ -287,48 +302,61 @@ static void
 rb_threadptr_exec_event_hooks_orig(rb_trace_arg_t *trace_arg, int pop_p)
 {
     rb_thread_t *th = trace_arg->th;
-    if (th->trace_arg == 0 &&
-	trace_arg->self != rb_mRubyVMFrozenCore /* skip special methods. TODO: remove it. */) {
-	const int vm_tracing = th->vm->trace_running;
-	const VALUE errinfo = th->errinfo;
-	const int outer_state = th->state;
-	int state = 0;
-	th->state = 0;
 
-	th->vm->trace_running++;
-	th->trace_arg = trace_arg;
-	{
-	    rb_hook_list_t *list;
-
-	    /* thread local traces */
-	    list = &th->event_hooks;
-	    if (list->events & trace_arg->event) {
-		state = exec_hooks(th, list, trace_arg, TRUE);
-		if (state) goto terminate;
-	    }
-
-	    /* vm global traces */
-	    list = &th->vm->event_hooks;
-	    if (list->events & trace_arg->event) {
-		state = exec_hooks(th, list, trace_arg, !vm_tracing);
-		if (state) goto terminate;
-	    }
-	    th->errinfo = errinfo;
+    if (trace_arg->event & RUBY_INTERNAL_EVENT_MASK) {
+	if (th->ec.trace_arg && (th->ec.trace_arg->event & RUBY_INTERNAL_EVENT_MASK)) {
+	    /* skip hooks because this thread doing INTERNAL_EVENT */
 	}
-      terminate:
-	th->trace_arg = 0;
-	th->vm->trace_running--;
+	else {
+	    rb_trace_arg_t *prev_trace_arg = th->ec.trace_arg;
+	    th->vm->trace_running++;
+	    th->ec.trace_arg = trace_arg;
+	    exec_hooks_unprotected(th, &th->event_hooks, trace_arg);
+	    exec_hooks_unprotected(th, &th->vm->event_hooks, trace_arg);
+	    th->ec.trace_arg = prev_trace_arg;
+	    th->vm->trace_running--;
+	}
+    }
+    else {
+	if (th->ec.trace_arg == NULL && /* check reentrant */
+	    trace_arg->self != rb_mRubyVMFrozenCore /* skip special methods. TODO: remove it. */) {
+	    const VALUE errinfo = th->ec.errinfo;
+	    const VALUE old_recursive = th->ec.local_storage_recursive_hash;
+	    int state = 0;
 
-	if (state) {
-	    if (pop_p) {
-		if (VM_FRAME_TYPE_FINISH_P(th->cfp)) {
-		    th->tag = th->tag->prev;
+	    th->ec.local_storage_recursive_hash = th->ec.local_storage_recursive_hash_for_trace;
+	    th->ec.errinfo = Qnil;
+
+	    th->vm->trace_running++;
+	    th->ec.trace_arg = trace_arg;
+	    {
+		/* thread local traces */
+		state = exec_hooks_protected(th, &th->event_hooks, trace_arg);
+		if (state) goto terminate;
+
+		/* vm global traces */
+		state = exec_hooks_protected(th, &th->vm->event_hooks, trace_arg);
+		if (state) goto terminate;
+
+		th->ec.errinfo = errinfo;
+	    }
+	  terminate:
+	    th->ec.trace_arg = NULL;
+	    th->vm->trace_running--;
+
+	    th->ec.local_storage_recursive_hash_for_trace = th->ec.local_storage_recursive_hash;
+	    th->ec.local_storage_recursive_hash = old_recursive;
+
+	    if (state) {
+		if (pop_p) {
+		    if (VM_FRAME_FINISHED_P(th->ec.cfp)) {
+			th->ec.tag = th->ec.tag->prev;
+		    }
+		    rb_vm_pop_frame(th);
 		}
-		th->cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(th->cfp);
+		TH_JUMP_TAG(th, state);
 	    }
-	    TH_JUMP_TAG(th, state);
 	}
-	th->state = outer_state;
     }
 }
 
@@ -348,22 +376,20 @@ VALUE
 rb_suppress_tracing(VALUE (*func)(VALUE), VALUE arg)
 {
     volatile int raised;
-    volatile int outer_state;
     VALUE result = Qnil;
-    rb_thread_t *th = GET_THREAD();
-    int state;
-    const int tracing = th->trace_arg ? 1 : 0;
+    rb_thread_t *volatile th = GET_THREAD();
+    enum ruby_tag_type state;
+    const int tracing = th->ec.trace_arg ? 1 : 0;
     rb_trace_arg_t dummy_trace_arg;
+    dummy_trace_arg.event = 0;
 
     if (!tracing) th->vm->trace_running++;
-    if (!th->trace_arg) th->trace_arg = &dummy_trace_arg;
+    if (!th->ec.trace_arg) th->ec.trace_arg = &dummy_trace_arg;
 
     raised = rb_threadptr_reset_raised(th);
-    outer_state = th->state;
-    th->state = 0;
 
     TH_PUSH_TAG(th);
-    if ((state = TH_EXEC_TAG()) == 0) {
+    if ((state = TH_EXEC_TAG()) == TAG_NONE) {
 	result = (*func)(arg);
     }
     TH_POP_TAG();
@@ -372,14 +398,16 @@ rb_suppress_tracing(VALUE (*func)(VALUE), VALUE arg)
 	rb_threadptr_set_raised(th);
     }
 
-    if (th->trace_arg == &dummy_trace_arg) th->trace_arg = 0;
+    if (th->ec.trace_arg == &dummy_trace_arg) th->ec.trace_arg = 0;
     if (!tracing) th->vm->trace_running--;
 
     if (state) {
-	JUMP_TAG(state);
+#if defined RUBY_USE_SETJMPEX && RUBY_USE_SETJMPEX
+	RB_GC_GUARD(result);
+#endif
+	TH_JUMP_TAG(th, state);
     }
 
-    th->state = outer_state;
     return result;
 }
 
@@ -413,8 +441,8 @@ static void call_trace_func(rb_event_flag_t, VALUE data, VALUE self, ID id, VALU
  *  +c-call+:: call a C-language routine
  *  +c-return+:: return from a C-language routine
  *  +call+:: call a Ruby method
- *  +class+:: start a class or module definition),
- *  +end+:: finish a class or module definition),
+ *  +class+:: start a class or module definition
+ *  +end+:: finish a class or module definition
  *  +line+:: execute code on a new line
  *  +raise+:: raise an exception
  *  +return+:: return from a Ruby method
@@ -486,10 +514,7 @@ thread_add_trace_func(rb_thread_t *th, VALUE trace)
 static VALUE
 thread_add_trace_func_m(VALUE obj, VALUE trace)
 {
-    rb_thread_t *th;
-
-    GetThreadPtr(obj, th);
-    thread_add_trace_func(th, trace);
+    thread_add_trace_func(rb_thread_ptr(obj), trace);
     return trace;
 }
 
@@ -505,19 +530,19 @@ thread_add_trace_func_m(VALUE obj, VALUE trace)
  */
 
 static VALUE
-thread_set_trace_func_m(VALUE obj, VALUE trace)
+thread_set_trace_func_m(VALUE target_thread, VALUE trace)
 {
-    rb_thread_t *th;
+    rb_thread_t *target_th = rb_thread_ptr(target_thread);
 
-    GetThreadPtr(obj, th);
-    rb_threadptr_remove_event_hook(th, call_trace_func, Qundef);
+    rb_threadptr_remove_event_hook(target_th, call_trace_func, Qundef);
 
     if (NIL_P(trace)) {
 	return Qnil;
     }
-
-    thread_add_trace_func(th, trace);
-    return trace;
+    else {
+	thread_add_trace_func(target_th, trace);
+	return trace;
+    }
 }
 
 static const char *
@@ -556,6 +581,7 @@ get_event_id(rb_event_flag_t event)
 	C(b_return, B_RETURN);
 	C(thread_begin, THREAD_BEGIN);
 	C(thread_end, THREAD_END);
+	C(fiber_switch, FIBER_SWITCH);
 	C(specified_line, SPECIFIED_LINE);
       case RUBY_EVENT_LINE | RUBY_EVENT_SPECIFIED_LINE: CONST_ID(id, "line"); return id;
 #undef C
@@ -567,15 +593,15 @@ get_event_id(rb_event_flag_t event)
 static void
 call_trace_func(rb_event_flag_t event, VALUE proc, VALUE self, ID id, VALUE klass)
 {
-    const char *srcfile = rb_sourcefile();
+    int line;
+    const char *srcfile = rb_source_loc(&line);
     VALUE eventname = rb_str_new2(get_event_name(event));
     VALUE filename = srcfile ? rb_str_new2(srcfile) : Qnil;
     VALUE argv[6];
-    int line = rb_sourceline();
     rb_thread_t *th = GET_THREAD();
 
     if (!klass) {
-	rb_thread_method_id_and_class(th, &id, &klass);
+	rb_thread_method_id_and_class(th, &id, 0, &klass);
     }
 
     if (klass) {
@@ -603,28 +629,20 @@ static VALUE rb_cTracePoint;
 
 typedef struct rb_tp_struct {
     rb_event_flag_t events;
+    int tracing; /* bool */
     rb_thread_t *target_th;
     void (*func)(VALUE tpval, void *data);
     void *data;
     VALUE proc;
-    int tracing;
     VALUE self;
 } rb_tp_t;
 
 static void
 tp_mark(void *ptr)
 {
-    if (ptr) {
-	rb_tp_t *tp = (rb_tp_t *)ptr;
-	rb_gc_mark(tp->proc);
-	if (tp->target_th) rb_gc_mark(tp->target_th->self);
-    }
-}
-
-static void
-tp_free(void *ptr)
-{
-    /* do nothing */
+    rb_tp_t *tp = ptr;
+    rb_gc_mark(tp->proc);
+    if (tp->target_th) rb_gc_mark(tp->target_th->self);
 }
 
 static size_t
@@ -635,7 +653,8 @@ tp_memsize(const void *ptr)
 
 static const rb_data_type_t tp_data_type = {
     "tracepoint",
-    {tp_mark, tp_free, tp_memsize,},
+    {tp_mark, RUBY_TYPED_NEVER_FREE, tp_memsize,},
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static VALUE
@@ -648,8 +667,12 @@ tp_alloc(VALUE klass)
 static rb_event_flag_t
 symbol2event_flag(VALUE v)
 {
-    static ID id;
-    VALUE sym = rb_convert_type(v, T_SYMBOL, "Symbol", "to_sym");
+    ID id;
+    VALUE sym = rb_convert_type_with_id(v, T_SYMBOL, "Symbol", idTo_sym);
+    const rb_event_flag_t RUBY_EVENT_A_CALL =
+	RUBY_EVENT_CALL | RUBY_EVENT_B_CALL | RUBY_EVENT_C_CALL;
+    const rb_event_flag_t RUBY_EVENT_A_RETURN =
+	RUBY_EVENT_RETURN | RUBY_EVENT_B_RETURN | RUBY_EVENT_C_RETURN;
 
 #define C(name, NAME) CONST_ID(id, #name); if (sym == ID2SYM(id)) return RUBY_EVENT_##NAME
     C(line, LINE);
@@ -664,9 +687,12 @@ symbol2event_flag(VALUE v)
     C(b_return, B_RETURN);
     C(thread_begin, THREAD_BEGIN);
     C(thread_end, THREAD_END);
+    C(fiber_switch, FIBER_SWITCH);
     C(specified_line, SPECIFIED_LINE);
+    C(a_call, A_CALL);
+    C(a_return, A_RETURN);
 #undef C
-    rb_raise(rb_eArgError, "unknown event: %s", rb_id2name(SYM2ID(sym)));
+    rb_raise(rb_eArgError, "unknown event: %"PRIsVALUE, rb_sym2str(sym));
 }
 
 static rb_tp_t *
@@ -680,7 +706,7 @@ tpptr(VALUE tpval)
 static rb_trace_arg_t *
 get_trace_arg(void)
 {
-    rb_trace_arg_t *trace_arg = GET_THREAD()->trace_arg;
+    rb_trace_arg_t *trace_arg = GET_THREAD()->ec.trace_arg;
     if (trace_arg == 0) {
 	rb_raise(rb_eRuntimeError, "access from outside");
     }
@@ -712,7 +738,7 @@ fill_path_and_lineno(rb_trace_arg_t *trace_arg)
 	rb_control_frame_t *cfp = rb_vm_get_ruby_level_next_cfp(trace_arg->th, trace_arg->cfp);
 
 	if (cfp) {
-	    trace_arg->path = cfp->iseq->location.path;
+	    trace_arg->path = rb_iseq_path(cfp->iseq);
 	    trace_arg->lineno = rb_vm_get_sourceline(cfp);
 	}
 	else {
@@ -740,7 +766,7 @@ fill_id_and_klass(rb_trace_arg_t *trace_arg)
 {
     if (!trace_arg->klass_solved) {
 	if (!trace_arg->klass) {
-	    rb_vm_control_frame_id_and_class(trace_arg->cfp, &trace_arg->id, &trace_arg->klass);
+	    rb_vm_control_frame_id_and_class(trace_arg->cfp, &trace_arg->id, &trace_arg->called_id, &trace_arg->klass);
 	}
 
 	if (trace_arg->klass) {
@@ -763,6 +789,13 @@ rb_tracearg_method_id(rb_trace_arg_t *trace_arg)
     return trace_arg->id ? ID2SYM(trace_arg->id) : Qnil;
 }
 
+VALUE
+rb_tracearg_callee_id(rb_trace_arg_t *trace_arg)
+{
+    fill_id_and_klass(trace_arg);
+    return trace_arg->called_id ? ID2SYM(trace_arg->called_id) : Qnil;
+}
+
 VALUE
 rb_tracearg_defined_class(rb_trace_arg_t *trace_arg)
 {
@@ -777,7 +810,7 @@ rb_tracearg_binding(rb_trace_arg_t *trace_arg)
     cfp = rb_vm_get_binding_creatable_next_cfp(trace_arg->th, trace_arg->cfp);
 
     if (cfp) {
-	return rb_binding_new_with_cfp(trace_arg->th, cfp);
+	return rb_vm_make_binding(trace_arg->th, cfp);
     }
     else {
 	return Qnil;
@@ -865,7 +898,7 @@ tracepoint_attr_path(VALUE tpval)
 }
 
 /*
- * Return the name of the method being called
+ * Return the name at the definition of the method being called
  */
 static VALUE
 tracepoint_attr_method_id(VALUE tpval)
@@ -873,6 +906,15 @@ tracepoint_attr_method_id(VALUE tpval)
     return rb_tracearg_method_id(get_trace_arg());
 }
 
+/*
+ * Return the called name of the method being called
+ */
+static VALUE
+tracepoint_attr_callee_id(VALUE tpval)
+{
+    return rb_tracearg_callee_id(get_trace_arg());
+}
+
 /*
  * Return class or module of the method being called.
  *
@@ -1135,12 +1177,43 @@ tracepoint_new(VALUE klass, rb_thread_t *target_th, rb_event_flag_t events, void
     return tpval;
 }
 
+/*
+ * Creates a tracepoint by registering a callback function for one or more
+ * tracepoint events. Once the tracepoint is created, you can use
+ * rb_tracepoint_enable to enable the tracepoint.
+ *
+ * Parameters:
+ *   1. VALUE target_thval - Meant for picking the thread in which the tracepoint
+ *      is to be created. However, current implementation ignore this parameter,
+ *      tracepoint is created for all threads. Simply specify Qnil.
+ *   2. rb_event_flag_t events - Event(s) to listen to.
+ *   3. void (*func)(VALUE, void *) - A callback function.
+ *   4. void *data - Void pointer that will be passed to the callback function.
+ *
+ * When the callback function is called, it will be passed 2 parameters:
+ *   1)VALUE tpval - the TracePoint object from which trace args can be extracted.
+ *   2)void *data - A void pointer which helps to share scope with the callback function.
+ *
+ * It is important to note that you cannot register callbacks for normal events and internal events
+ * simultaneously because they are different purpose.
+ * You can use any Ruby APIs (calling methods and so on) on normal event hooks.
+ * However, in internal events, you can not use any Ruby APIs (even object creations).
+ * This is why we can't specify internal events by TracePoint directly.
+ * Limitations are MRI version specific.
+ *
+ * Example:
+ *   rb_tracepoint_new(Qnil, RUBY_INTERNAL_EVENT_NEWOBJ | RUBY_INTERNAL_EVENT_FREEOBJ, obj_event_i, data);
+ *
+ *   In this example, a callback function obj_event_i will be registered for
+ *   internal events RUBY_INTERNAL_EVENT_NEWOBJ and RUBY_INTERNAL_EVENT_FREEOBJ.
+ */
 VALUE
 rb_tracepoint_new(VALUE target_thval, rb_event_flag_t events, void (*func)(VALUE, void *), void *data)
 {
-    rb_thread_t *target_th = 0;
+    rb_thread_t *target_th = NULL;
+
     if (RTEST(target_thval)) {
-	GetThreadPtr(target_thval, target_th);
+	target_th = rb_thread_ptr(target_thval);
 	/* TODO: Test it!
 	 * Warning: This function is not tested.
 	 */
@@ -1237,7 +1310,7 @@ static VALUE
 tracepoint_inspect(VALUE self)
 {
     rb_tp_t *tp = tpptr(self);
-    rb_trace_arg_t *trace_arg = GET_THREAD()->trace_arg;
+    rb_trace_arg_t *trace_arg = GET_THREAD()->ec.trace_arg;
 
     if (trace_arg) {
 	switch (trace_arg->event) {
@@ -1246,7 +1319,7 @@ tracepoint_inspect(VALUE self)
 	    {
 		VALUE sym = rb_tracearg_method_id(trace_arg);
 		if (NIL_P(sym))
-		  goto default_inspect;
+		    goto default_inspect;
 		return rb_sprintf("#<TracePoint:%"PRIsVALUE"@%"PRIsVALUE":%d in `%"PRIsVALUE"'>",
 				  rb_tracearg_event(trace_arg),
 				  rb_tracearg_path(trace_arg),
@@ -1280,6 +1353,50 @@ tracepoint_inspect(VALUE self)
     }
 }
 
+static void
+tracepoint_stat_event_hooks(VALUE hash, VALUE key, rb_event_hook_t *hook)
+{
+    int active = 0, deleted = 0;
+
+    while (hook) {
+	if (hook->hook_flags & RUBY_EVENT_HOOK_FLAG_DELETED) {
+	    deleted++;
+	}
+	else {
+	    active++;
+	}
+	hook = hook->next;
+    }
+
+    rb_hash_aset(hash, key, rb_ary_new3(2, INT2FIX(active), INT2FIX(deleted)));
+}
+
+/*
+ * call-seq:
+ *	TracePoint.stat -> obj
+ *
+ *  Returns internal information of TracePoint.
+ *
+ *  The contents of the returned value are implementation specific.
+ *  It may be changed in future.
+ *
+ *  This method is only for debugging TracePoint itself.
+ */
+
+static VALUE
+tracepoint_stat_s(VALUE self)
+{
+    rb_vm_t *vm = GET_VM();
+    VALUE stat = rb_hash_new();
+
+    tracepoint_stat_event_hooks(stat, vm->self, vm->event_hooks.hooks);
+    /* TODO: thread local hooks */
+
+    return stat;
+}
+
+static void Init_postponed_job(void);
+
 /* This function is called from inits.c */
 void
 Init_vm_trace(void)
@@ -1333,11 +1450,11 @@ Init_vm_trace(void)
      * +:b_return+:: event hook at block ending
      * +:thread_begin+:: event hook at thread beginning
      * +:thread_end+:: event hook at thread ending
+     * +:fiber_switch+:: event hook at fiber switch
      *
      */
     rb_cTracePoint = rb_define_class("TracePoint", rb_cObject);
     rb_undef_alloc_func(rb_cTracePoint);
-    rb_undef_method(CLASS_OF(rb_cTracePoint), "new");
     rb_define_singleton_method(rb_cTracePoint, "new", tracepoint_new_s, -1);
     /*
      * Document-method: trace
@@ -1365,67 +1482,137 @@ Init_vm_trace(void)
     rb_define_method(rb_cTracePoint, "lineno", tracepoint_attr_lineno, 0);
     rb_define_method(rb_cTracePoint, "path", tracepoint_attr_path, 0);
     rb_define_method(rb_cTracePoint, "method_id", tracepoint_attr_method_id, 0);
+    rb_define_method(rb_cTracePoint, "callee_id", tracepoint_attr_callee_id, 0);
     rb_define_method(rb_cTracePoint, "defined_class", tracepoint_attr_defined_class, 0);
     rb_define_method(rb_cTracePoint, "binding", tracepoint_attr_binding, 0);
     rb_define_method(rb_cTracePoint, "self", tracepoint_attr_self, 0);
     rb_define_method(rb_cTracePoint, "return_value", tracepoint_attr_return_value, 0);
     rb_define_method(rb_cTracePoint, "raised_exception", tracepoint_attr_raised_exception, 0);
+
+    rb_define_singleton_method(rb_cTracePoint, "stat", tracepoint_stat_s, 0);
+
+    /* initialized for postponed job */
+
+    Init_postponed_job();
 }
 
 typedef struct rb_postponed_job_struct {
-    unsigned long flags; /* reserve */
-    rb_thread_t *th; /* created thread, reserve */
+    unsigned long flags; /* reserved */
+    struct rb_thread_struct *th; /* created thread, reserved */
     rb_postponed_job_func_t func;
     void *data;
-    struct rb_postponed_job_struct *next;
 } rb_postponed_job_t;
 
-int
-rb_postponed_job_register(unsigned int flags, rb_postponed_job_func_t func, void *data)
+#define MAX_POSTPONED_JOB                  1000
+#define MAX_POSTPONED_JOB_SPECIAL_ADDITION   24
+
+static void
+Init_postponed_job(void)
 {
-    rb_thread_t *th = GET_THREAD();
-    rb_vm_t *vm = th->vm;
-    rb_postponed_job_t *pjob = (rb_postponed_job_t *)ruby_xmalloc(sizeof(rb_postponed_job_t));
-    if (pjob == NULL) return 0; /* failed */
+    rb_vm_t *vm = GET_VM();
+    vm->postponed_job_buffer = ALLOC_N(rb_postponed_job_t, MAX_POSTPONED_JOB);
+    vm->postponed_job_index = 0;
+}
+
+enum postponed_job_register_result {
+    PJRR_SUCESS      = 0,
+    PJRR_FULL        = 1,
+    PJRR_INTERRUPTED = 2
+};
+
+static enum postponed_job_register_result
+postponed_job_register(rb_thread_t *th, rb_vm_t *vm,
+		       unsigned int flags, rb_postponed_job_func_t func, void *data, int max, int expected_index)
+{
+    rb_postponed_job_t *pjob;
+
+    if (expected_index >= max) return PJRR_FULL; /* failed */
+
+    if (ATOMIC_CAS(vm->postponed_job_index, expected_index, expected_index+1) == expected_index) {
+	pjob = &vm->postponed_job_buffer[expected_index];
+    }
+    else {
+	return PJRR_INTERRUPTED;
+    }
 
     pjob->flags = flags;
     pjob->th = th;
     pjob->func = func;
     pjob->data = data;
 
-    pjob->next = vm->postponed_job;
-    vm->postponed_job = pjob;
-
     RUBY_VM_SET_POSTPONED_JOB_INTERRUPT(th);
-    return 1;
+
+    return PJRR_SUCESS;
 }
 
+
+/* return 0 if job buffer is full */
+int
+rb_postponed_job_register(unsigned int flags, rb_postponed_job_func_t func, void *data)
+{
+    rb_thread_t *th = GET_THREAD();
+    rb_vm_t *vm = th->vm;
+
+  begin:
+    switch (postponed_job_register(th, vm, flags, func, data, MAX_POSTPONED_JOB, vm->postponed_job_index)) {
+      case PJRR_SUCESS     : return 1;
+      case PJRR_FULL       : return 0;
+      case PJRR_INTERRUPTED: goto begin;
+      default: rb_bug("unreachable\n");
+    }
+}
+
+/* return 0 if job buffer is full */
 int
 rb_postponed_job_register_one(unsigned int flags, rb_postponed_job_func_t func, void *data)
 {
-    rb_vm_t *vm = GET_VM();
-    rb_postponed_job_t *pjob = vm->postponed_job;
+    rb_thread_t *th = GET_THREAD();
+    rb_vm_t *vm = th->vm;
+    rb_postponed_job_t *pjob;
+    int i, index;
 
-    while (pjob) {
+  begin:
+    index = vm->postponed_job_index;
+    for (i=0; i<index; i++) {
+	pjob = &vm->postponed_job_buffer[i];
 	if (pjob->func == func) {
+	    RUBY_VM_SET_POSTPONED_JOB_INTERRUPT(th);
 	    return 2;
 	}
-	pjob = pjob->next;
     }
-
-    return rb_postponed_job_register(flags, func, data);
+    switch (postponed_job_register(th, vm, flags, func, data, MAX_POSTPONED_JOB + MAX_POSTPONED_JOB_SPECIAL_ADDITION, index)) {
+      case PJRR_SUCESS     : return 1;
+      case PJRR_FULL       : return 0;
+      case PJRR_INTERRUPTED: goto begin;
+      default: rb_bug("unreachable\n");
+    }
 }
 
 void
 rb_postponed_job_flush(rb_vm_t *vm)
 {
-    rb_postponed_job_t *pjob = vm->postponed_job, *next_pjob;
-    vm->postponed_job = 0;
+    rb_thread_t *th = GET_THREAD();
+    const unsigned long block_mask = POSTPONED_JOB_INTERRUPT_MASK|TRAP_INTERRUPT_MASK;
+    unsigned long saved_mask = th->interrupt_mask & block_mask;
+    VALUE saved_errno = th->ec.errinfo;
 
-    while (pjob) {
-	next_pjob = pjob->next;
-	pjob->func(pjob->data);
-	ruby_xfree(pjob);
-	pjob = next_pjob;
+    th->ec.errinfo = Qnil;
+    /* mask POSTPONED_JOB dispatch */
+    th->interrupt_mask |= block_mask;
+    {
+	TH_PUSH_TAG(th);
+	if (EXEC_TAG() == TAG_NONE) {
+	    int index;
+	    while ((index = vm->postponed_job_index) > 0) {
+		if (ATOMIC_CAS(vm->postponed_job_index, index, index-1) == index) {
+		    rb_postponed_job_t *pjob = &vm->postponed_job_buffer[index-1];
+		    (*pjob->func)(pjob->data);
+		}
+	    }
+	}
+	TH_POP_TAG();
     }
+    /* restore POSTPONED_JOB mask */
+    th->interrupt_mask &= ~(saved_mask ^ block_mask);
+    th->ec.errinfo = saved_errno;
 }
